const std = @import("std");
const fs = std.fs;
const io = std.io;
const os = std.os;
const fmt = std.fmt;
const mem = std.mem;
const heap = std.heap;
const meta = std.meta;
const ArrayList = std.ArrayList;
const Allocator = std.mem.Allocator;

const flags = @import("flags.zig");

const DB = @import("DB.zig");
const File = @import("File.zig");
const PackedUser = @import("PackedUser.zig");
const Header = @import("header.zig").Header;
const section_length_bits = @import("header.zig").section_length_bits;

const usage =
    \\usage: turbo-analyze [options] [db.turbo]
    \\
    \\  -h          Print this help message and exit
    \\  [db.turbo]  Path to the turbonss database file (default: db.turbo)
    \\
;

const Info = struct {
    fname: []const u8,
    size_file: []const u8,
    bytes_file: os.off_t,
    version: meta.fieldInfo(Header, .version).field_type,
    endian: []const u8,
    ptr_size: meta.fieldInfo(Header, .ptr_size).field_type,
    getgr_bufsize: meta.fieldInfo(Header, .getgr_bufsize).field_type,
    getpw_bufsize: meta.fieldInfo(Header, .getpw_bufsize).field_type,
    users: meta.fieldInfo(Header, .num_users).field_type,
    groups: meta.fieldInfo(Header, .num_groups).field_type,
    shells: meta.fieldInfo(Header, .num_shells).field_type,
};

pub fn main() !void {
    // This line is here because of https://github.com/ziglang/zig/issues/7807
    const argv: []const [*:0]const u8 = os.argv;

    const stderr = io.getStdErr().writer();
    const stdout = io.getStdOut().writer();

    const return_code = execute(stdout, stderr, argv[1..]);
    os.exit(return_code);
}

fn execute(
    stdout: anytype,
    stderr: anytype,
    argv: []const [*:0]const u8,
) u8 {
    const myflags = flags.parse(argv, &[_]flags.Flag{
        .{ .name = "-h", .kind = .boolean },
    }) catch {
        stderr.writeAll(usage) catch {};
        return 1;
    };

    if (myflags.boolFlag("-h")) {
        stdout.writeAll(usage) catch return 1;
        return 0;
    }

    const db_file = switch (myflags.args.len) {
        0 => "db.turbo",
        1 => mem.span(myflags.args[0]),
        else => {
            stderr.print("ERROR: too many arguments\n", .{}) catch {};
            stderr.writeAll(usage) catch {};
            return 1;
        },
    };

    const file_size_bytes = blk: {
        const fd = os.open(db_file, os.O.RDONLY, 0) catch |err| {
            stderr.print("ERROR: failed to open '{s}': {s}\n", .{
                db_file,
                @errorName(err),
            }) catch {};
            return 1;
        };
        defer os.close(fd);
        const stat = os.fstat(fd) catch |err| {
            stderr.print("ERROR: fstat '{s}': {s}\n", .{
                db_file,
                @errorName(err),
            }) catch {};
            return 1;
        };
        break :blk stat.size;
    };

    var file = File.open(db_file) catch |err| {
        stderr.print(
            "ERROR {s}: file '{s}' is corrupted or cannot be read\n",
            .{ @errorName(err), db_file },
        ) catch {};
        return 1;
    };
    defer file.close();
    const db = file.db;

    var file_size_scratch: [16]u8 = undefined;
    const buf = fmt.bufPrint(file_size_scratch[0..], "{:.2}", .{
        fmt.fmtIntSizeBin(@intCast(u64, file_size_bytes)),
    }) catch unreachable;

    const info = Info{
        .fname = db_file,
        .size_file = buf,
        .bytes_file = file_size_bytes,
        .version = db.header.version,
        .endian = @tagName(db.header.endian),
        .ptr_size = db.header.ptr_size,
        .getgr_bufsize = db.header.getgr_bufsize,
        .getpw_bufsize = db.header.getpw_bufsize,
        .users = db.header.num_users,
        .groups = db.header.num_groups,
        .shells = db.header.num_shells,
    };

    const template =
        \\File:               {[fname]s}
        \\Size:               {[size_file]s} ({[bytes_file]d} bytes)
        \\Version:            {[version]d}
        \\Endian:             {[endian]s}
        \\Pointer size:       {[ptr_size]} bytes
        \\getgr buffer size:  {[getgr_bufsize]d}
        \\getpw buffer size:  {[getpw_bufsize]d}
        \\Users:              {[users]d}
        \\Groups:             {[groups]d}
        \\Shells:             {[shells]d}
        \\Sections:
        \\    Name                 Begin    End            Size
        \\
    ;
    stdout.print(template, info) catch {};

    var lengths = DB.fieldLengths(db.header);
    var offsets = DB.fieldOffsets(lengths);

    inline for (meta.fields(DB.DBNumbers)) |field| {
        const length = @field(lengths, field.name);
        const size = fmt.bufPrint(file_size_scratch[0..], "{:.0}", .{
            fmt.fmtIntSizeBin(@intCast(u64, length << section_length_bits)),
        }) catch unreachable;

        const start = @field(offsets, field.name);
        const end = start + @field(lengths, field.name);
        stdout.print("    {s:<21}{x:0>8} {x:0>8} {s:>10}\n", .{
            field.name,
            start << section_length_bits,
            end << section_length_bits,
            size,
        }) catch {};
    }

    return 0;
}

const testing = std.testing;

test "trivial error: db file" {
    const args = &[_][*:0]const u8{};
    const allocator = testing.allocator;
    var stderr = ArrayList(u8).init(allocator);
    defer stderr.deinit();
    var stdout = ArrayList(u8).init(allocator);
    defer stdout.deinit();

    const exit_code = execute(stdout.writer(), stderr.writer(), args[0..]);
    try testing.expectEqual(@as(u8, 1), exit_code);
    try testing.expectEqualStrings(
        stderr.items,
        "ERROR: failed to open 'db.turbo': FileNotFound\n",
    );
}
